# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import absolute_import, division, print_function

import numpy
import xarray
from affine import Affine
from datacube.utils import geometry
from dateutil.parser import parse
from ows.util import Version
from rasterio import MemoryFile

from datacube_ows.cube_pool import cube
from datacube_ows.data import DataStacker
from datacube_ows.mv_index import MVSelectOpts
from datacube_ows.ogc_exceptions import WCS1Exception
from datacube_ows.ogc_utils import ConfigException
from datacube_ows.ows_configuration import get_config
from datacube_ows.resource_limits import ResourceLimited


class WCS1GetCoverageRequest():
    version = Version(1, 0, 0)
    # pylint: disable=too-many-instance-attributes, too-many-branches, too-many-statements, too-many-locals
    def __init__(self, args):
        self.args = args
        cfg = get_config()

        # Argument: Coverage (required)
        if "coverage" not in args:
            raise WCS1Exception("No coverage specified",
                                WCS1Exception.MISSING_PARAMETER_VALUE,
                                locator="COVERAGE parameter",
                                valid_keys=list(cfg.product_index))
        self.product_name = args["coverage"]
        self.product = cfg.product_index.get(self.product_name)
        if not self.product or not self.product.wcs:
            raise WCS1Exception("Invalid coverage: %s" % self.product_name,
                                WCS1Exception.COVERAGE_NOT_DEFINED,
                                locator="COVERAGE parameter",
                                valid_keys=list(cfg.product_index))

        # Argument: FORMAT (required)
        if "format" not in args:
            raise WCS1Exception("No FORMAT parameter supplied",
                                WCS1Exception.MISSING_PARAMETER_VALUE,
                                locator="FORMAT parameter",
                                valid_keys=cfg.wcs_formats_by_name)
        if args["format"] not in cfg.wcs_formats_by_name:
            raise WCS1Exception("Unsupported format: %s" % args["format"],
                                WCS1Exception.INVALID_PARAMETER_VALUE,
                                locator="FORMAT parameter",
                                valid_keys=cfg.wcs_formats_by_name)
        self.format = cfg.wcs_formats_by_name[args["format"]]

        # Argument: (request) CRS (required)
        if "crs" not in args:
            raise WCS1Exception("No request CRS specified",
                                WCS1Exception.MISSING_PARAMETER_VALUE,
                                locator="CRS parameter",
                                valid_keys=list(cfg.published_CRSs))
        self.request_crsid = args["crs"]
        if self.request_crsid not in cfg.published_CRSs:
            raise WCS1Exception("%s is not a supported CRS" % self.request_crsid,
                                WCS1Exception.INVALID_PARAMETER_VALUE,
                                locator="CRS parameter",
                                valid_keys=list(cfg.published_CRSs))
        self.request_crs = cfg.crs(self.request_crsid)

        # Argument: response_crs (optional)
        if "response_crs" in args:
            self.response_crsid = args["response_crs"]
            if self.response_crsid not in cfg.published_CRSs:
                raise WCS1Exception("%s is not a supported CRS" % self.response_crsid,
                                    WCS1Exception.INVALID_PARAMETER_VALUE,
                                    locator="RESPONSE_CRS parameter",
                                    valid_keys=list(cfg.published_CRSs))
            self.response_crs = geometry.CRS(self.response_crsid)
            self.response_crs = cfg.crs(self.response_crsid)
        else:
            self.response_crsid = self.request_crsid
            self.response_crs = self.request_crs

        # Arguments: One of BBOX or TIME is required
        # if "bbox" not in args and "time" not in args:
        #    raise WCS1Exception("At least one of BBOX or TIME parameters must be supplied",
        #                        WCS1Exception.MISSING_PARAMETER_VALUE,
        #                        locator="BBOX or TIME parameter"
        #                        )

        # Argument: BBOX (technically not required if TIME supplied, but
        #       it's not clear to me what that would mean.)
        # For WCS 1.0.0 all bboxes will be specified as minx, miny, maxx, maxy
        if "bbox" not in args:
            raise WCS1Exception("No BBOX parameter supplied",
                                WCS1Exception.MISSING_PARAMETER_VALUE,
                                locator="BBOX or TIME parameter")
        try:
            self.minx, self.miny, self.maxx, self.maxy = map(float, args['bbox'].split(','))
        except:
            raise WCS1Exception("Invalid BBOX parameter",
                                WCS1Exception.INVALID_PARAMETER_VALUE,
                                locator="BBOX parameter")

        # Argument: TIME
        # if self.product.wcs_sole_time:
        #    self.times = [parse(self.product.wcs_sole_time).date()]
        if "time" not in args:
            #      CEOS treats no supplied time argument as all time.
            # I'm really not sure what the right thing to do is, but QGIS wants us to do SOMETHING - use configured
            # default.
            self.times = [self.product.default_time]
        else:
            # TODO: the min/max/res format option?
            # It's a bit underspeced. I'm not sure what the "res" would look like.
            times = args["time"].split(",")
            self.times = []
            for t in times:
                if t == "now":
                    continue
                try:
                    time = parse(t).date()
                    if time not in self.product.ranges["time_set"]:
                        raise WCS1Exception(
                            "Time value '%s' not a valid date for coverage %s" % (t, self.product_name),
                            WCS1Exception.INVALID_PARAMETER_VALUE,
                            locator="TIME parameter",
                            valid_keys=[d.strftime('%Y-%m-%d') for d in self.product.ranges["time_set"]]
                        )
                    self.times.append(time)
                except ValueError:
                    raise WCS1Exception(
                        "Time value '%s' not a valid ISO-8601 date" % t,
                        WCS1Exception.INVALID_PARAMETER_VALUE,
                        locator="TIME parameter",
                        valid_keys=[d.strftime('%Y-%m-%d') for d in self.product.ranges["time_set"]]
                    )
            self.times.sort()

            if len(self.times) == 0:
                raise WCS1Exception(
                    "No valid ISO-8601 dates",
                    WCS1Exception.INVALID_PARAMETER_VALUE,
                    locator="TIME parameter",
                    valid_keys = [d.strftime('%Y-%m-%d') for d in self.product.ranges["time_set"]]
                )
            elif len(self.times) > 1 and not self.format["multi-time"]:
                raise WCS1Exception(
                    "Cannot select more than one time slice with the %s format" % self.format["name"],
                    WCS1Exception.INVALID_PARAMETER_VALUE,
                    locator="TIME and FORMAT parameters"
                )

        # Range constraint parameter: MEASUREMENTS
        # No default is set in the DescribeCoverage, so it is required
        # But QGIS wants us to work without one, so take default from config
        if "measurements" in args:
            bands = args["measurements"]
            self.bands = []
            for b in bands.split(","):
                if not b:
                    continue
                try:
                    self.bands.append(self.product.band_idx.locale_band(b))
                except ConfigException:
                    raise WCS1Exception(f"Invalid measurement: {b}",
                                        WCS1Exception.INVALID_PARAMETER_VALUE,
                                        locator="MEASUREMENTS parameter",
                                        valid_keys=self.product.band_idx.band_labels())
            if not bands:
                raise WCS1Exception("No measurements supplied",
                                    WCS1Exception.INVALID_PARAMETER_VALUE,
                                    locator="MEASUREMENTS parameter",
                                    valid_keys = self.product.band_idx.band_labels())
        elif "styles" in args and args["styles"]:
            # Use style bands.
            # Non-standard protocol extension.
            #
            # As we have correlated WCS and WMS service implementations,
            # we can accept a style from WMS, and return the bands used for it.
            styles = args["styles"].split(",")
            if len(styles) != 1:
                raise WCS1Exception("Multiple style parameters not supported")
            style = self.product.style_index.get(styles[0])
            if style:
                self.bands = set()
                for b in style.needed_bands:
                    if b not in style.flag_bands:
                        self.bands.add(b)
            else:
                self.bands = self.product.wcs_default_bands
        else:
            self.bands = self.product.wcs_default_bands

        # Argument: EXCEPTIONS (optional - defaults to XML)
        if "exceptions" in args and args["exceptions"] != "application/vnd.ogc.se_xml":
            raise WCS1Exception(f"Unsupported exception format: {args['exceptions']}",
                                WCS1Exception.INVALID_PARAMETER_VALUE,
                                locator="EXCEPTIONS parameter")

        # Argument: INTERPOLATION (optional only nearest-neighbour currently supported.)
        #      If 'none' is supported in future, validation of width/height/res will need to change.
        if "interpolation" in args and args["interpolation"] != "nearest neighbor":
            raise WCS1Exception(f'Unsupported interpolation method: {args["interpolation"]}',
                                WCS1Exception.INVALID_PARAMETER_VALUE,
                                locator="INTERPOLATION parameter")

        if "width" in args:
            if "resx" in args or "resy" in args:
                raise WCS1Exception("Specify WIDTH/HEIGHT parameters OR RESX/RESY parameters - not both",
                                    WCS1Exception.MISSING_PARAMETER_VALUE,
                                    locator="RESX/RESY/WIDTH/HEIGHT parameters")
            if "height" not in args:
                raise WCS1Exception("WIDTH parameter supplied without HEIGHT parameter",
                                    WCS1Exception.MISSING_PARAMETER_VALUE,
                                    locator="WIDTH/HEIGHT parameters")
            try:
                self.height = int(args["height"])
                if self.height < 1:
                    raise ValueError()
            except ValueError:
                raise WCS1Exception("HEIGHT parameter must be a positive integer",
                                    WCS1Exception.INVALID_PARAMETER_VALUE,
                                    locator="HEIGHT parameter")
            try:
                self.width = int(args["width"])
                if self.width < 1:
                    raise ValueError()
            except ValueError:
                raise WCS1Exception("WIDTH parameter must be a positive integer",
                                    WCS1Exception.INVALID_PARAMETER_VALUE,
                                    locator="WIDTH parameter")
            self.resx = (self.maxx - self.minx) / self.width
            self.resy = (self.maxy - self.miny) / self.height
        elif "resx" in args:
            if "height" in args:
                raise WCS1Exception("Specify WIDTH/HEIGHT parameters OR RESX/RESY parameters - not both",
                                    WCS1Exception.MISSING_PARAMETER_VALUE,
                                    locator="RESX/RESY/WIDTH/HEIGHT parameters")
            if "resy" not in args:
                raise WCS1Exception("RESX parameter supplied without RESY parameter",
                                    WCS1Exception.MISSING_PARAMETER_VALUE,
                                    locator="RESX/RESY parameters")
            try:
                self.resx = float(args["resx"])
                if self.resx <= 0.0:
                    raise ValueError(0)
            except ValueError:
                raise WCS1Exception("RESX parameter must be a positive number",
                                    WCS1Exception.INVALID_PARAMETER_VALUE,
                                    locator="RESX parameter")
            try:
                self.resy = float(args["resy"])
                if self.resy <= 0.0:
                    raise ValueError(0)
            except ValueError:
                raise WCS1Exception("RESY parameter must be a positive number",
                                    WCS1Exception.INVALID_PARAMETER_VALUE,
                                    locator="RESY parameter")
            self.width = (self.maxx - self.minx) / self.resx
            self.height = (self.maxy - self.miny) / self.resy
            self.width = int(self.width + 0.5)
            self.height = int(self.height + 0.5)
        elif "height" in args:
            raise WCS1Exception("HEIGHT parameter supplied without WIDTH parameter",
                                WCS1Exception.MISSING_PARAMETER_VALUE,
                                locator="WIDTH/HEIGHT parameters")
        elif "resy" in args:
            raise WCS1Exception("RESY parameter supplied without RESX parameter",
                                WCS1Exception.MISSING_PARAMETER_VALUE,
                                locator="RESX/RESY parameters")
        else:
            raise WCS1Exception("You must specify either the WIDTH/HEIGHT parameters or RESX/RESY",
                                WCS1Exception.MISSING_PARAMETER_VALUE,
                                locator="RESX/RESY/WIDTH/HEIGHT parameters")

        self.extent = geometry.polygon([(self.minx, self.miny),
                                        (self.minx, self.maxy),
                                        (self.maxx, self.maxy),
                                        (self.maxx, self.miny),
                                        (self.minx, self.miny)
                                       ],
                                       self.request_crs
                                      )

        xscale = (self.maxx - self.minx) / self.width
        yscale = (self.miny - self.maxy) / self.height
        trans_aff = Affine.translation(self.minx, self.maxy)
        scale_aff = Affine.scale(xscale, yscale)
        self.affine = trans_aff * scale_aff
        self.geobox = geometry.GeoBox(self.width, self.height, self.affine, self.request_crs)


def get_coverage_data(req):
    # pylint: disable=too-many-locals, protected-access
    with cube() as dc:
        if not dc:
            raise WCS1Exception("Database connectivity failure")
        stacker = DataStacker(req.product,
                              req.geobox,
                              req.times,
                              bands=req.bands)
        n_datasets = stacker.datasets(dc.index, mode=MVSelectOpts.COUNT)
        if n_datasets == 0:
            # Return an empty coverage file with full metadata?
            cfg = get_config()
            x_range = (req.minx, req.maxx)
            y_range = (req.miny, req.maxy)
            xname = cfg.published_CRSs[req.request_crsid]["horizontal_coord"]
            yname = cfg.published_CRSs[req.request_crsid]["vertical_coord"]
            xvals = numpy.linspace(
                x_range[0],
                x_range[1],
                num=req.width
            )
            yvals = numpy.linspace(
                y_range[0],
                y_range[1],
                num=req.height
            )
            if cfg.published_CRSs[req.request_crsid]["vertical_coord_first"]:
                nparrays = {
                    band: (("time", yname, xname),
                           numpy.full((len(req.times), len(yvals), len(xvals)),
                                      req.product.band_idx.nodata_val(band))
                          )
                    for band in req.bands
                }
            else:
                nparrays = {
                    band: (("time", xname, yname),
                           numpy.full((len(req.times), len(xvals), len(yvals)),
                                      req.product.band_idx.nodata_val(band))
                          )
                    for band in req.bands
                }
            data = xarray.Dataset(
                nparrays,
                coords={
                    "time": req.times,
                    xname: xvals,
                    yname: yvals,
                }
            ).astype("int16")

            return n_datasets, data

        try:
            req.product.resource_limits.check_wcs(n_datasets)
        except ResourceLimited as e:
            raise WCS1Exception(
                f"This request processes too much data to be served in a reasonable amount of time. ({e}) "
                + "Please reduce the bounds of your request and try again.")
        datasets = stacker.datasets(index=dc.index)
        output = stacker.data(datasets, skip_corrections=True)

        # Clean extent flag band from output
        for k, v in output.data_vars.items():
            if k not in req.bands:
                output = output.drop_vars([k])
        return n_datasets, output


def get_tiff(req, data):
    """Uses rasterio MemoryFiles in order to return a streamable GeoTiff response"""
    # Copied from CEOS.  Does not seem to support multi-time dimension data - is this even possible in GeoTiff?
    supported_dtype_map = {
        'uint8': 1,
        'uint16': 2,
        'int16': 3,
        'uint32': 4,
        'int32': 5,
        'float32': 6,
        'float64': 7,
        'complex': 9,
        'complex64': 10,
        'complex128': 11,
    }

    dtype_list = [data[array].dtype for array in data.data_vars]
    dtype = str(max(dtype_list, key=lambda d: supported_dtype_map[str(d)]))

    data = data.squeeze(dim="time", drop=True)
    data = data.astype(dtype)
    cfg = get_config()
    xname = cfg.published_CRSs[req.request_crsid]["horizontal_coord"]
    yname = cfg.published_CRSs[req.request_crsid]["vertical_coord"]
    nodata = 0
    for band in data.data_vars:
        nodata = req.product.band_idx.nodata_val(band)
    with MemoryFile() as memfile:
        # pylint: disable=protected-access, bad-continuation
        with memfile.open(
            driver="GTiff",
            width=data.dims[xname],
            height=data.dims[yname],
            count=len(data.data_vars),
            transform=req.affine,
            crs=req.response_crsid,
            nodata=nodata,
            tiled=True,
            compress="lzw",
            interleave="band",
            dtype=dtype) as dst:
            for idx, band in enumerate(data.data_vars, start=1):
                dst.write(data[band].values, idx)
                dst.set_band_description(idx, req.product.band_idx.band_label(band))
                if cfg.wcs_tiff_statistics:
                    dst.update_tags(idx, STATISTICS_MINIMUM=data[band].values.min())
                    dst.update_tags(idx, STATISTICS_MAXIMUM=data[band].values.max())
                    dst.update_tags(idx, STATISTICS_MEAN=data[band].values.mean())
                    dst.update_tags(idx, STATISTICS_STDDEV=data[band].values.std())
        return memfile.read()


def get_netcdf(req, data):
    # Cleanup dataset attributes for NetCDF export
    data.attrs["crs"] = req.response_crsid # geometry.CRS(response_crs)
    for k, v in data.data_vars.items():
        v.attrs["crs"] = req.response_crsid
        if "spectral_definition" in v.attrs:
            del v.attrs["spectral_definition"]
        if "flags_definition" in v.attrs:
            del v.attrs["flags_definition"]
    if "time" in data and "units" in data["time"].attrs:
        del data["time"].attrs["units"]

    # And export to NetCDF
    return data.to_netcdf()



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
from collections import defaultdict
from decimal import ROUND_HALF_UP, Decimal
from math import isclose

import matplotlib
import numpy
from colour import Color
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap, to_hex
from xarray import Dataset

matplotlib.use('Agg')

from datacube_ows.ogc_utils import ConfigException, FunctionWrapper
from datacube_ows.styles.base import StyleDefBase
from datacube_ows.styles.expression import Expression

_LOG = logging.getLogger(__name__)

UNSCALED_DEFAULT_RAMP = [
    {
        "value": -1e-24,
        "color": "#000080",
        "alpha": 0.0
    },
    {
        "value": 0.0,
        "color": "#000080",
    },
    {
        "value": 0.1,
        "color": "#0000FF",
    },
    {
        "value": 0.3,
        "color": "#00FFFF",
    },
    {
        "value": 0.5,
        "color": "#00FF00",
    },
    {
        "value": 0.7,
        "color": "#FFFF00",
    },
    {
        "value": 0.9,
        "color": "#FF0000",
    },
    {
        "value": 1.0,
        "color": "#800000",
    },
]


def scale_unscaled_ramp(rmin, rmax, unscaled):
    if isinstance(rmin, str):
        rmin = float(rmin)
    if isinstance(rmax, str):
        rmax = float(rmax)
    return [
        {
            "value": (rmax - rmin) * u["value"] + rmin,
            "color": u["color"],
            "alpha": u.get("alpha", 1.0)
        } for u in unscaled
    ]


def crack_ramp(ramp):
    values = []
    red = []
    green = []
    blue = []
    alpha = []
    for r in ramp:
        values.append(float(r["value"]))
        color = Color(r["color"])
        red.append(color.red)
        green.append(color.green)
        blue.append(color.blue)
        alpha.append(r.get("alpha", 1.0))

    return (values, red, green, blue, alpha)


def read_mpl_ramp(mpl_ramp: str):
    unscaled_cmap = []
    try:
        cmap = plt.get_cmap(mpl_ramp)
    except:
        raise ConfigException(f"Invalid Matplotlib name: {mpl_ramp}")
    val_range = numpy.arange(0.1, 1.1, 0.1)
    rgba_hex = to_hex(cmap(0.0))
    unscaled_cmap.append(
        {
            "value": 0.0,
            "color": rgba_hex,
            "alpha": 1.0
        }
    )
    for val in val_range:
        rgba_hex = to_hex(cmap(val))
        unscaled_cmap.append(
            {
                "value": float(val),
                "color": rgba_hex
            }
        )
    return unscaled_cmap


def colour_ramp_legend(bytesio, legend_cfg, colour_ramp, map_name,
                       default_title):
    if colour_ramp.legend_legacy:
        return legacy_colour_ramp_legend(bytesio, legend_cfg,
                                  colour_ramp, map_name,
                                  default_title)

    def create_cdict_ticks(cfg, ramp):
        normalize_factor = ramp.legend_end - ramp.legend_begin

        cdict = dict()
        bands = defaultdict(list)
        started = False
        finished = False
        for index, ramp_point in enumerate(ramp.ramp):
            if finished:
                continue

            value = ramp_point.get("value")
            normalized = (value - float(ramp.legend_begin)) / float(normalize_factor)

            if not started:
                if isclose(value, float(ramp.legend_begin), abs_tol=1e-9):
                    started = True
                else:
                    continue
            if not finished:
                if isclose(value, float(ramp.legend_end), abs_tol=1e-9):
                    finished = True

            for band, intensity in ramp.components.items():
                bands[band].append((normalized, intensity[index], intensity[index]))

        for band, blist in bands.items():
            cdict[band] = tuple(blist)

        ticks = dict()
        for tick, tick_lbl in zip(ramp.ticks, ramp.tick_labels):
            value = float(tick)
            normalized = (value - float(ramp.legend_begin)) / float(normalize_factor)
            ticks[normalized] = tick_lbl

        if len(ticks) == 0:
            ticks = None
        return cdict, ticks

    cdict, ticks = create_cdict_ticks(legend_cfg, colour_ramp)

    plt.rcdefaults()
    if colour_ramp.legend_mpl_rcparams:
        plt.rcParams.update(colour_ramp.legend_mpl_rcparams)
    fig = plt.figure(figsize=(colour_ramp.legend_width,
                              colour_ramp.legend_height))
    ax = fig.add_axes(colour_ramp.legend_strip_location)
    custom_map = LinearSegmentedColormap(map_name, cdict)
    color_bar = matplotlib.colorbar.ColorbarBase(
        ax,
        cmap=custom_map,
        orientation="horizontal")

    if ticks is not None:
        color_bar.set_ticks(list(ticks.keys()))
        color_bar.set_ticklabels(list(ticks.values()))

    title = colour_ramp.legend_title if colour_ramp.legend_title else default_title
    if colour_ramp.legend_units:
        title = title + "(" + colour_ramp.legend_units + ")"

    color_bar.set_label(title)

    plt.savefig(bytesio, format='png')


def legacy_colour_ramp_legend(bytesio, legend_cfg, colour_ramp, map_name,
                       default_title):
    def custom_label(label, custom_config):
        prefix = custom_config.get("prefix", "")
        l = custom_config.get("label", label)
        suffix = custom_config.get("suffix", "")
        return f"{prefix}{l}{suffix}"

    # Create custom cdict for matplotlib colorramp
    # Matplot lib color dicts must start at 0 and end at 1
    # because of this we normalize the values
    # Also create tick labels based on configuration
    # ticks are also normalized between 0 - 1.0
    # so they are position correctly on the colorbar
    def create_cdict_ticks(cfg, ramp):
        generate = (cfg.get("major_ticks", None) is not None or
                    cfg.get("scale_by", None) is not None or
                    cfg.get("radix_point", None) is not None)

        return from_definition(cfg, ramp, generate)

    def find_clc(ramp, last=False):
        l = ramp.ramp if not last else reversed(ramp.ramp)
        for index, value in enumerate(l):
            fwd_index = index if not last else (len(ramp.ramp) - (index + 1))
            if "legend" in value:
                return fwd_index
        return 0 if not last else (len(ramp.ramp) - 1)

    def from_definition(cfg, ramp, generate):
        tick_mod = cfg.get("major_ticks", 1)
        tick_offset = cfg.get("offset", 0)
        tick_scale = cfg.get("scale_by", 1)
        places = cfg.get("radix_point", 1)

        start_index = find_clc(ramp) if not generate else 0
        stop_index = find_clc(ramp, last=True) if not generate else (len(ramp.ramp) - 1)

        start = ramp.ramp[start_index].get("value")
        stop = ramp.ramp[stop_index].get("value")
        normalize_factor = stop - start

        ticks = dict()
        cdict = dict()
        bands = defaultdict(list)
        for index, ramp_val in enumerate(ramp.ramp):
            if index < start_index or index > stop_index:
                continue

            value = ramp_val.get("value")
            normalized = (value - start) / normalize_factor
            custom_legend_cfg = ramp_val.get("legend", None)

            mod_close = False
            mod_equal = False
            if generate:
                mod_close = isclose((value * tick_scale + tick_offset) % (tick_mod * tick_scale), 0.0, abs_tol=1e-8)
                mod_equal = value % tick_mod == 0

            if mod_close or mod_equal:
                label = value * tick_scale + tick_offset
                label = round(label, places) if places > 0 else int(label)
                ticks[normalized] = label

            if custom_legend_cfg is not None:
                label = custom_label(value, custom_legend_cfg)
                ticks[normalized] = label

            for band, intensity in ramp.components.items():
                bands[band].append((normalized, intensity[index], intensity[index]))

        for band, blist in bands.items():
            cdict[band] = tuple(blist)

        if len(ticks) == 0:
            ticks = None
        return cdict, ticks

    cdict, ticks = create_cdict_ticks(legend_cfg, colour_ramp)

    # TODO: Potentially short-circuit this if using string based mpl_ramp
    # custom_map = plt.get_cmap(style_cfg["mpl_ramp"]) should return a LinearSegmentedColormap

    plt.rcdefaults()
    if legend_cfg.get("rcParams", None) is not None:
        plt.rcParams.update(legend_cfg.get("rcParams"))
    fig = plt.figure(figsize=(legend_cfg.get("width", 4),
                              legend_cfg.get("height", 1.25)))
    ax_pos = legend_cfg.get("axes_position", [0.05, 0.5, 0.9, 0.15])
    ax = fig.add_axes(ax_pos)
    custom_map = LinearSegmentedColormap(map_name, cdict)
    color_bar = matplotlib.colorbar.ColorbarBase(
        ax,
        cmap=custom_map,
        orientation="horizontal")

    if ticks is not None:
        color_bar.set_ticks(list(ticks.keys()))
        color_bar.set_ticklabels([str(l) for l in ticks.values()])

    title = legend_cfg.get("title", default_title)
    unit = legend_cfg.get("units", "unitless")
    title = title + "(" + unit + ")"

    color_bar.set_label(title)

    plt.savefig(bytesio, format='png')


class ColorRamp:
    def __init__(self, style, ramp_cfg):
        self.style = style
        if "color_ramp" in ramp_cfg:
            raw_scaled_ramp = ramp_cfg["color_ramp"]
        else:
            rmin, rmax = ramp_cfg["range"]
            unscaled_ramp = UNSCALED_DEFAULT_RAMP
            if "mpl_ramp" in ramp_cfg:
                unscaled_ramp = read_mpl_ramp(ramp_cfg["mpl_ramp"])
            raw_scaled_ramp = scale_unscaled_ramp(
                rmin, rmax, unscaled_ramp)
        self.ramp = raw_scaled_ramp
        legend_cfg = ramp_cfg.get("legend", {})
        if legend_cfg.get("show_legend", True) and not legend_cfg.get("url"):
            self.parse_legend(legend_cfg)
        else:
            self.auto_legend = False

        self.crack_ramp()

        if self.auto_legend and not self.legend_legacy:
            fbegin = float(self.legend_begin)
            fend = float(self.legend_end)
            begin_in_ramp = False
            end_in_ramp = False
            begin_before_idx = None
            end_before_idx = None
            for idx, col_point in enumerate(self.ramp):
                col_val = col_point["value"]
                if not begin_in_ramp and begin_before_idx is None:
                    if isclose(col_val, fbegin, abs_tol=1e-9):
                        begin_in_ramp = True
                    elif col_val > fbegin:
                        begin_before_idx = idx
                if not end_in_ramp and end_before_idx is None:
                    if isclose(col_val, fend, abs_tol=1e-9):
                        end_in_ramp = True
                    elif col_val > fend:
                        end_before_idx = idx
            if not begin_in_ramp:
                color, alpha = self.color_alpha_at(fbegin)
                begin_col_point = {
                    "value": fbegin,
                    "color": color.get_hex(),
                    "alpha": alpha
                }
                if begin_before_idx is None:
                    self.ramp.append(begin_col_point)
                else:
                    self.ramp.insert(begin_before_idx, begin_col_point)
                if end_before_idx is not None:
                    end_before_idx += 1
            if not end_in_ramp:
                color, alpha = self.color_alpha_at(fend)
                end_col_point = {
                    "value": fend,
                    "color": color.get_hex(),
                    "alpha": alpha
                }
                if end_before_idx is None:
                    self.ramp.append(end_col_point)
                else:
                    self.ramp.insert(end_before_idx, end_col_point)
            if not end_in_ramp or not begin_in_ramp:
                self.crack_ramp()

    def crack_ramp(self):
        values, r, g, b, a = crack_ramp(self.ramp)
        self.values = values
        self.components = {
            "red": r,
            "green": g,
            "blue": b,
            "alpha": a
        }

    def parse_legend(self, cfg):
        def rounder_str(prec):
            rstr = "1"
            if prec == 0:
                return rstr
            rstr += "."
            for i in range(prec - 1):
                rstr += "0"
            rstr += "1"
            return rstr

        self.auto_legend = True
        self.legend_title = cfg.get("title")
        self.legend_units = cfg.get("units", "")
        self.legend_decimal_places = cfg.get("decimal_places", 1)
        if self.legend_decimal_places < 0:
            raise ConfigException("decimal_places cannot be negative")
        self.rounder = Decimal(rounder_str(self.legend_decimal_places))
        self.parse_legend_range(cfg)
        self.parse_legend_ticks(cfg)
        self.parse_legend_tick_labels(cfg)
        self.parse_legend_matplotlib_args(cfg)
        self.legend_legacy = any(
            legent in cfg
            for legent in ["major_ticks", "offset", "scale_by", "radix_point"]
        )
        if not self.legend_legacy and all(
            legent not in cfg
            for legent in ["begin", "end", "decimal_places", "ticks_every", "tick_count", "tick_labels", "ticks"]
        ):
            # No legacy entries, but no new entries either.
            # Check ramp for legend tips
            for r in self.ramp:
                if "legend" in r:
                    self.legend_legacy = True
                    break
        if self.legend_legacy:
            _LOG.warning("Style %s uses deprecated legend configuration.  Please refer to the documentation and update your config",
                         self.style.name)

    def parse_legend_range(self, cfg):
        # pylint: disable=attribute-defined-outside-init
        self.legend_begin = None
        # pylint: disable=attribute-defined-outside-init
        self.legend_end = None
        if "begin" in cfg:
            self.legend_begin = Decimal(cfg["begin"])
        else:
            for col_def in self.ramp:
                if isclose(col_def.get("alpha", 1.0), 1.0, abs_tol=1e-9):
                    self.legend_begin = Decimal(col_def["value"])
                    break
            if self.legend_begin is None:
                self.legend_begin = Decimal(self.ramp[0]["value"])
        if "end" in cfg:
            self.legend_end = Decimal(cfg["end"])
        else:
            for col_def in reversed(self.ramp):
                if col_def.get("alpha", 1.0) == 1.0:
                    self.legend_end = Decimal(col_def["value"])
                    break
            if self.legend_end is None:
                self.legend_end = Decimal(self.ramp[-1]["value"])

    def parse_legend_ticks(self, cfg):
        # Ticks
        ticks = []
        if "ticks_every" in cfg:
            if "tick_count" in cfg:
                raise ConfigException("Cannot use tick count and ticks_every in the same legend")
            if "ticks" in cfg:
                raise ConfigException("Cannot use ticks and ticks_every in the same legend")
            delta = Decimal(cfg["ticks_every"])
            tickval = self.legend_begin
            while tickval < self.legend_end:
                ticks.append(tickval)
                tickval += delta
            ticks.append(tickval)
        if "ticks" in cfg:
            if "tick_count" in cfg:
                raise ConfigException("Cannot use tick count and ticks in the same legend")
            ticks = [Decimal(t) for t in cfg["ticks"]]
        if not ticks:
            count = int(cfg.get("tick_count", 1))
            if count < 0:
                raise ConfigException("tick_count cannot be negative")
            delta = self.legend_end - self.legend_begin
            dcount = Decimal(count)

            for i in range(0, count + 1):
                tickval = self.legend_begin + (Decimal(i) / dcount) * delta
                ticks.append(tickval.quantize(self.rounder, rounding=ROUND_HALF_UP))
        # pylint: disable=attribute-defined-outside-init
        self.ticks = ticks

    def parse_legend_tick_labels(self, cfg):
        labels = cfg.get("tick_labels", {})
        defaults = labels.get("default", {})
        default_prefix = defaults.get("prefix", "")
        default_suffix = defaults.get("suffix", "")
        # pylint: disable=attribute-defined-outside-init
        self.tick_labels = []
        for tick in self.ticks:
            label_cfg = labels.get(str(tick))
            if label_cfg:
                prefix = label_cfg.get("prefix", default_prefix)
                suffix = label_cfg.get("suffix", default_suffix)
                label  = label_cfg.get("label", str(tick))
                self.tick_labels.append(prefix + label + suffix)
            else:
                self.tick_labels.append(
                    default_prefix + str(tick) + default_suffix
                )
    def parse_legend_matplotlib_args(self, cfg):
        # pylint: disable=attribute-defined-outside-init
        self.legend_mpl_rcparams = cfg.get("rcParams", {})
        self.legend_width = cfg.get("width", 4)
        self.legend_height = cfg.get("height", 1.25)
        self.legend_strip_location = cfg.get("strip_location",
                                      [0.05, 0.5, 0.9, 0.15])

    def get_value(self, data, band):
        return numpy.interp(data, self.values, self.components[band])

    def get_8bit_value(self, data, band):
        val = self.get_value(data, band)
        return (val * 255).astype("uint8")

    def apply(self, data):
        imgdata = {}
        for band in self.components:
            imgdata[band] = (data.dims, self.get_8bit_value(data, band))
        imgdataset = Dataset(imgdata, coords=data.coords)
        return imgdataset

    def color_alpha_at(self, val):
        color = Color(
            rgb=(
                self.get_value(val, "red").item(),
                self.get_value(val, "green").item(),
                self.get_value(val, "blue").item(),
            )
        )
        alpha = self.get_value(val, "alpha")

        return color, alpha


class ColorRampDef(StyleDefBase):
    auto_legend = True
    def __init__(self, product, style_cfg, stand_alone=False, defer_multi_date=False, user_defined=False):
        super(ColorRampDef, self).__init__(product, style_cfg,
                           stand_alone=stand_alone, defer_multi_date=True, user_defined=user_defined)
        style_cfg = self._raw_cfg
        self.color_ramp = ColorRamp(self, style_cfg)
        self.include_in_feature_info = style_cfg.get("include_in_feature_info", True)

        if "index_function" in style_cfg:
            self.index_function = FunctionWrapper(self,
                                                  style_cfg["index_function"],
                                                  stand_alone=self.stand_alone)
            if not self.stand_alone:
                for band in style_cfg["needed_bands"]:
                    self.raw_needed_bands.add(band)
        elif "index_expression" in style_cfg:
            self.index_function = Expression(self, style_cfg["index_expression"])
            for band in self.index_function.needed_bands:
                self.raw_needed_bands.add(band)
            if self.stand_alone:
                self.needed_bands = [self.local_band(b) for b in self.raw_needed_bands]
                self.flag_bands = []
        else:
            raise ConfigException("Index function is required for index and hybrid styles. Style %s in layer %s" % (
                self.name,
                self.product.name
            ))
        if not defer_multi_date:
            self.parse_multi_date(style_cfg)

    def apply_index(self, data):
        index_data = self.index_function(data)
        data['index_function'] = (index_data.dims, index_data.data)
        return data["index_function"]

    def transform_single_date_data(self, data):
        d = self.apply_index(data)
        return self.color_ramp.apply(d)

    def single_date_legend(self, bytesio):
        colour_ramp_legend(bytesio,
                           self.legend_cfg,
                           self.color_ramp,
                           self.product.name,
                           self.title
                           )

    class MultiDateHandler(StyleDefBase.MultiDateHandler):
        auto_legend = True
        def __init__(self, style, cfg):
            super().__init__(style, cfg)
            if self.animate:
                self.feature_info_label = None
                self.color_ramp = style.color_ramp
            else:
                self.feature_info_label = cfg.get("feature_info_label", None)
                self.color_ramp = ColorRamp(style, cfg)

        def transform_data(self, data):
            xformed_data = self.style.apply_index(data)
            agg = self.aggregator(xformed_data)
            return self.color_ramp.apply(agg)

        def legend(self, bytesio):
            if self.animate and not self.legend_cfg:
                return self.style.single_date_legend(bytesio)
            title = self.legend_cfg.get("title", self.range_str() + " Dates")
            name = self.style.product.name + f"_{self.min_count}"
            colour_ramp_legend(bytesio,
                               self.legend_cfg,
                               self.color_ramp,
                               name,
                               title
                               )
            return True

StyleDefBase.register_subclass(ColorRampDef,
                               ("range", "color_ramp")
)



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import absolute_import, division, print_function

import logging
import os
import warnings

import sentry_sdk
from datacube.utils.aws import configure_s3_access
from flask import Flask, request
from flask_babel import Babel
from flask_log_request_id import RequestID, RequestIDLogFilter
from prometheus_flask_exporter.multiprocess import \
    GunicornInternalPrometheusMetrics
from rasterio.errors import NotGeoreferencedWarning
from sentry_sdk.integrations.flask import FlaskIntegration

from datacube_ows.ows_configuration import get_config

__all__ = [
    'initialise_babel',
    'initialise_logger',
    'initialise_ignorable_warnings',
    'initialise_debugging',
    'initialise_sentry',
    'initialise_aws_credentials',
    'parse_config_file',
    'initialise_flask',
    'initialise_prometheus',
    'initialise_prometheus_register',
    'generate_locale_selector'
]



def initialise_logger(name=None):
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("[%(asctime)s] %(name)s [%(request_id)s] [%(levelname)s] %(message)s"))
    handler.addFilter(RequestIDLogFilter())
    _LOG = logging.getLogger(name)
    _LOG.addHandler(handler)
    # If invoked using Gunicorn, link our root logger to the gunicorn logger
    # this will mean the root logs will be captured and managed by the gunicorn logger
    # allowing you to set the gunicorn log directories and levels for logs
    # produced by this application
    _LOG.setLevel(logging.getLogger('gunicorn.error').getEffectiveLevel())
    return _LOG


def initialise_ignorable_warnings():
    # Suppress annoying rasterio warning message every time we write to a non-georeferenced image format
    warnings.simplefilter("ignore", category=NotGeoreferencedWarning)


def initialise_debugging(log=None):
    # PYCHARM Debugging
    if os.environ.get("PYDEV_DEBUG"):
        import pydevd_pycharm
        pydevd_pycharm.settrace('172.17.0.1', port=12321, stdoutToServer=True, stderrToServer=True)
        if log:
            log.info("PyCharm Debugging enabled")


def initialise_sentry(log=None):
    if os.environ.get("SENTRY_KEY") and os.environ.get("SENTRY_PROJECT"):
        SENTRY_ENV_TAG = os.environ.get("SENTRY_ENV_TAG") if os.environ.get("SENTRY_ENV_TAG") else "dev"
        sentry_sdk.init(
            dsn="https://%s@sentry.io/%s" % (os.environ["SENTRY_KEY"], os.environ["SENTRY_PROJECT"]),
            environment=SENTRY_ENV_TAG,
            integrations = [FlaskIntegration()]
        )
        if log:
            log.info("Sentry initialised")


def initialise_aws_credentials(log=None):
    # Startup initialisation of libraries controlled by environment variables
    #
    # Move to a function to facilitate unit testing.
    # Should be done in a more flexible pluggable way.

    # Boto3/AWS
    if os.environ.get("AWS_DEFAULT_REGION"):
        env_nosign = os.environ.get("AWS_NO_SIGN_REQUEST", "yes")
        unsigned = bool(env_nosign)
        if not unsigned or env_nosign.lower() in ("n", "f", "no", "false", "0"):
            unsigned = False
            # delete env variable
            del os.environ["AWS_NO_SIGN_REQUEST"]
        else:
            # Workaround for rasterio bug
            os.environ["AWS_ACCESS_KEY_ID"] = "fake"
            os.environ["AWS_SECRET_ACCESS_KEY"] = "fake"
        env_requester_pays = os.environ.get("AWS_REQUEST_PAYER", "")
        requester_pays = False
        if env_requester_pays.lower() == "requester":
            requester_pays = True
        if log:
            if unsigned:
                log.info("S3 access configured with unsigned requests")
            else:
                log.info("S3 access configured with signed requests")
        credentials = configure_s3_access(aws_unsigned=unsigned, requester_pays=requester_pays)

        if "AWS_S3_ENDPOINT" in os.environ and os.environ["AWS_S3_ENDPOINT"] == "":
            del os.environ["AWS_S3_ENDPOINT"]
    elif log:
        log.warning("Environment variable $AWS_DEFAULT_REGION not set.  (This warning can be ignored if all data is stored locally.)")


def parse_config_file(log=None):
    # Cache a parsed config file object
    # (unless deferring to first request)
    cfg = None
    if not os.environ.get("DEFER_CFG_PARSE"):
        cfg = get_config()
    return cfg


def initialise_flask(name):
    app = Flask(name.split('.')[0])
    RequestID(app)
    return app


def initialise_prometheus(app, log=None):
    # Prometheus
    if os.environ.get("prometheus_multiproc_dir", False):
        metrics = GunicornInternalPrometheusMetrics(app)
        if log:
            log.info("Prometheus metrics enabled")
        return metrics
    return None


def initialise_prometheus_register(metrics):
    # Register routes with Prometheus - call after all routes set up.
    if os.environ.get("prometheus_multiproc_dir", False):
        metrics.register_default(
            metrics.summary(
                'flask_ows_request_full_url', 'Request summary by request url',
                labels={
                    'query_request': lambda: request.args.get('request'),
                    'query_service': lambda: request.args.get('service'),
                    'query_layers': lambda: request.args.get('layers'),
                    'query_url': lambda: request.full_path
                }
            )
        )


def generate_locale_selector(locales):
    def selector_template():
        return request.accept_languages.best_match(locales)
    return selector_template

def initialise_babel(cfg, app):
    if cfg and cfg.internationalised:
        app.config["BABEL_TRANSLATION_DIRECTORIES"] = cfg.translations_dir
        babel = Babel(app,
                      default_locale=cfg.locales[0],
                      default_domain=cfg.message_domain,
                      configure_jinja=False
                      )
        babel.localeselector(generate_locale_selector(cfg.locales))
        return babel
    else:
        return None


# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

#pylint: skip-file

from __future__ import absolute_import, division, print_function

from datetime import datetime, timedelta, timezone

import datacube
from psycopg2.extras import Json

from datacube_ows.ogc_utils import NoTimezoneException, tz_for_coord
from datacube_ows.ows_configuration import \
    get_config  # , get_layers, ProductLayerDef
from datacube_ows.utils import get_sqlconn


def get_crsids(cfg=None):
    if not cfg:
        cfg = get_config()
    return cfg.internal_CRSs.keys()


def get_crses(cfg=None):
    return {crsid: datacube.utils.geometry.CRS(crsid) for crsid in get_crsids(cfg)}


def jsonise_bbox(bbox):
    if isinstance(bbox, dict):
        return bbox
    else:
        return {
            "top": bbox.top,
            "bottom": bbox.bottom,
            "left": bbox.left,
            "right": bbox.right,
        }


def create_multiprod_range_entry(dc, product, crses):
    print("Merging multiproduct ranges for %s (ODC products: %s)" % (
        product.name,
        repr(product.product_names)
    ))
    conn = get_sqlconn(dc)
    prodids = [p.id for p in product.products]
    wms_name = product.name

    if all(
            not datasets_exist(dc, p_name)
            for p_name in product.product_names
    ):
        print("Could not find datasets for any product in multiproduct: ", product.name)
        conn.close()
        print("Done")
        return

    txn = conn.begin()
    # Attempt to insert row
    conn.execute("""
        INSERT INTO wms.multiproduct_ranges
        (wms_product_name,lat_min,lat_max,lon_min,lon_max,dates,bboxes)
        VALUES
        (%(p_id)s, 0, 0, 0, 0, %(empty)s, %(empty)s)
        ON CONFLICT (wms_product_name) DO NOTHING
        """,
                 {"p_id": wms_name, "empty": Json("")})

    # Update extents
    conn.execute("""
        UPDATE wms.multiproduct_ranges
        SET lat_min = subq.lat_min,
            lat_max = subq.lat_max,
            lon_min = subq.lon_min,
            lon_max = subq.lon_max
        FROM (
            select min(lat_min) as lat_min,
                   max(lat_max) as lat_max,
                   min(lon_min) as lon_min,
                   max(lon_max) as lon_max
            from wms.product_ranges
            where id = ANY (%(p_prodids)s)
        ) as subq
        WHERE wms_product_name = %(p_id)s
        """,
                 {"p_id": wms_name, "p_prodids": prodids})

    # Create sorted list of dates
    results = conn.execute(
        """
        SELECT dates
        FROM   wms.product_ranges
        WHERE  id  = ANY (%(p_prodids)s)
        """, {"p_prodids": prodids}
    )
    dates = set()
    for r in results:
        for d in r[0]:
            dates.add(d)
    dates = sorted(dates)
    conn.execute("""
           UPDATE wms.multiproduct_ranges
           SET dates = %(dates)s
           WHERE wms_product_name= %(p_id)s
      """,
                 {
                     "dates": Json(dates),
                     "p_id": wms_name
                 }
    )

    # calculate bounding boxes
    results = list(conn.execute("""
        SELECT lat_min,lat_max,lon_min,lon_max
        FROM wms.multiproduct_ranges
        WHERE wms_product_name=%(p_id)s
        """,
        {"p_id": wms_name}))

    r = results[0]

    epsg4326 = datacube.utils.geometry.CRS("EPSG:4326")
    box = datacube.utils.geometry.box(
        float(r[2]),
        float(r[0]),
        float(r[3]),
        float(r[1]),
        epsg4326)

    cfg = get_config()
    conn.execute("""
        UPDATE wms.multiproduct_ranges
        SET bboxes = %s::jsonb
        WHERE wms_product_name=%s
        """,
                 Json({crsid: jsonise_bbox(box.to_crs(crs).boundingbox) for crsid, crs in get_crses(cfg).items()}),
                 wms_name
    )

    txn.commit()
    conn.close()
    return


def create_range_entry(dc, product, crses, summary_product=False):
  print("Updating range for ODC product %s..." % product.name)
  # NB. product is an ODC product
  conn = get_sqlconn(dc)
  txn = conn.begin()
  prodid = product.id

  # insert empty row if one does not already exist
  conn.execute("""
    INSERT INTO wms.product_ranges
    (id,lat_min,lat_max,lon_min,lon_max,dates,bboxes)
    VALUES
    (%(p_id)s, 0, 0, 0, 0, %(empty)s, %(empty)s)
    ON CONFLICT (id) DO NOTHING
    """,
    {"p_id": prodid, "empty": Json("")})


  # Update min/max lat/longs
  conn.execute(
      """
      UPDATE wms.product_ranges pr
      SET lat_min = st_ymin(subq.bbox),
          lat_max = st_ymax(subq.bbox),
          lon_min = st_xmin(subq.bbox),
          lon_max = st_xmax(subq.bbox)
      FROM (
        SELECT st_extent(stv.spatial_extent) as bbox
        FROM public.space_time_view stv
        WHERE stv.dataset_type_ref = %(p_id)s
      ) as subq
      WHERE pr.id = %(p_id)s
      """,
      {"p_id": prodid})

  # Set default timezone
  conn.execute("""
    set timezone to 'Etc/UTC'
  """)


  if summary_product:
      # Loop over dates
      dates = set()

      results = conn.execute(
          """
          select
                array_agg(temporal_extent)
          from public.space_time_view
          WHERE dataset_type_ref = %(p_id)s
          """,
          {"p_id": prodid}
      )
      for result in results:
          for dat_ran in result[0]:
              dates.add(dat_ran.lower)

      dates = sorted(dates)

      conn.execute("""
           UPDATE wms.product_ranges
           SET dates = %(dates)s
           WHERE id= %(p_id)s
      """,
                   {
                       "dates": Json([t.strftime("%Y-%m-%d") for t in dates]),
                       "p_id": prodid
                   }
      )
  else:
      # Loop over dates
      dates = set()

      results = conn.execute(
          """
          select
                lower(temporal_extent), upper(temporal_extent),
                ST_X(ST_Centroid(spatial_extent)),
                ST_Y(ST_Centroid(spatial_extent))
          from public.space_time_view
          WHERE dataset_type_ref = %(p_id)s
          """ %
          {"p_id": prodid}
      )
      for result in results:
          dt1, dt2, lon, lat = result
          try:
              tz = tz_for_coord(lon, lat)
          except NoTimezoneException:
              offset = round(lon / 15.0)
              tz = timezone(timedelta(hours=offset))
          dates.add(dt1.astimezone(tz).date())
      dates = sorted(dates)

      conn.execute("""
           UPDATE wms.product_ranges
           SET dates = %(dates)s
           WHERE id= %(p_id)s
      """,
                   {
                       "dates": Json([t.strftime("%Y-%m-%d") for t in dates]),
                       "p_id": prodid
                   }
                   )

  # calculate bounding boxes
  results = list(conn.execute("""
    SELECT lat_min,lat_max,lon_min,lon_max
    FROM wms.product_ranges
    WHERE id=%(p_id)s
    """,
      {"p_id": prodid}))

  r = results[0]

  epsg4326 = datacube.utils.geometry.CRS("EPSG:4326")
  box = datacube.utils.geometry.box(
    float(r[2]),
    float(r[0]),
    float(r[3]),
    float(r[1]),
    epsg4326)

  conn.execute("""
    UPDATE wms.product_ranges
    SET bboxes = %(bbox)s::jsonb
    WHERE id=%(p_id)s
    """, {
    "bbox": Json(
      {crsid: {"top": box.to_crs(crs).boundingbox.top,
               "bottom": box.to_crs(crs).boundingbox.bottom,
               "left": box.to_crs(crs).boundingbox.left,
               "right": box.to_crs(crs).boundingbox.right}
        for crsid, crs in crses.items()
       }
    ),
    "p_id": product.id})

  txn.commit()
  conn.close()


def datasets_exist(dc, product_name):
  conn = get_sqlconn(dc)

  results = conn.execute("""
    SELECT COUNT(*)
    FROM agdc.dataset ds, agdc.dataset_type p
    WHERE ds.archived IS NULL
    AND ds.dataset_type_ref = p.id
    AND p.name = %s""",
    product_name)

  conn.close()

  return list(results)[0][0] > 0


def add_ranges(dc, product_names, summary=False, merge_only=False):
    odc_products = {}
    ows_multiproducts = []
    errors = False
    for pname in product_names:
        dc_product = None
        ows_product = get_config().product_index.get(pname)
        if not ows_product:
            ows_product = get_config().native_product_index.get(pname)
        if ows_product:
            for dc_pname in ows_product.product_names:
                if dc_pname in odc_products:
                    odc_products[dc_pname]["ows"].append(ows_product)
                else:
                    odc_products[dc_pname] = {"ows": [ows_product]}
            print("OWS Layer %s maps to ODC Product(s): %s" % (
                ows_product.name,
                repr(ows_product.product_names)
            ))
            if ows_product.multi_product:
                ows_multiproducts.append(ows_product)
        if not ows_product:
            dc_product = dc.index.products.get_by_name(pname)
            if dc_product:
                print("ODC Layer: %s" % pname)
                if pname in odc_products:
                    odc_products[pname]["ows"].append(None)
                else:
                    odc_products[pname] = {"ows": [None]}
            else:
                print("Unrecognised product name:", pname)
                errors = True
                continue

    if ows_multiproducts and merge_only:
        print("Merge-only: Skipping range update of products:", repr(list(odc_products.keys())))
    else:
        for pname, ows_prods in odc_products.items():
            dc_product = dc.index.products.get_by_name(pname)
            if dc_product is None:
                print("Could not find ODC product:", pname)
                errors = True
            elif datasets_exist(dc, dc_product.name):
                prod_summary = summary
                for ows_prod in ows_prods["ows"]:
                    if ows_prod:
                        prod_summary = not ows_prod.is_raw_time_res
                        break
                create_range_entry(dc, dc_product, get_crses(), prod_summary)
            else:
                print("Could not find any datasets for: ", pname)
    for mp in ows_multiproducts:
        create_multiprod_range_entry(dc, mp, get_crses())

    print("Done.")
    return errors

def get_ranges(dc, product, path=None, is_dc_product=False):
    cfg = product.global_cfg
    conn = get_sqlconn(dc)
    if not is_dc_product and product.multi_product:
        if path is not None:
            raise Exception("Combining subproducts and multiproducts is not yet supported")
        results = conn.execute("""
            SELECT *
            FROM wms.multiproduct_ranges
            WHERE wms_product_name=%s""",
                               product.name
                              )
    else:
        if is_dc_product:
            prod_id = product.id
        else:
            prod_id = product.product.id
        if path is not None:
            results = conn.execute("""
                SELECT *
                FROM wms.sub_product_ranges
                WHERE product_id=%s and sub_product_id=%s""",
                                   prod_id, path
                                  )
        else:
            results = conn.execute("""
                SELECT *
                FROM wms.product_ranges
                WHERE id=%s""",
                                   prod_id
                                  )
    for result in results:
        conn.close()
        times = [datetime.strptime(d, "%Y-%m-%d").date() for d in result["dates"] if d is not None]
        if not times:
            return None
        return {
            "lat": {
                "min": float(result["lat_min"]),
                "max": float(result["lat_max"]),
            },
            "lon": {
                "min": float(result["lon_min"]),
                "max": float(result["lon_max"]),
            },
            "times": times,
            "start_time": times[0],
            "end_time": times[-1],
            "time_set": set(times),
            "bboxes": cfg.alias_bboxes(result["bboxes"])
        }
    return None



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
from datetime import datetime

import numpy
from colour import Color
from datacube.utils.masking import make_mask
from matplotlib import patches as mpatches
from matplotlib import pyplot as plt
from xarray import DataArray, Dataset, merge

from datacube_ows.config_utils import ConfigException, OWSConfigEntry
from datacube_ows.styles.base import StyleDefBase

_LOG = logging.getLogger(__name__)


class ValueMapRule(OWSConfigEntry):
    def __init__(self, style_def, band, cfg):
        super().__init__(self, cfg)
        self.style = style_def
        self.band = style_def.local_band(band)

        self.title = cfg["title"]
        self.abstract = cfg.get("abstract")
        if self.title and self.abstract:
            self.label = f"{self.title} - {self.abstract}"
        elif self.title:
            self.label = self.title
        elif self.abstract:
            self.label = self.abstract
        else:
            self.label = None
        self.color_str = cfg["color"]
        self.rgb = Color(self.color_str)
        if cfg.get("mask", False):
            self.alpha = 0.0
        else:
            self.alpha = cfg.get("alpha", 1.0)

        if "flags" in cfg:
            flags = cfg["flags"]
            self.or_flags = False
            if "or" in flags:
                self.or_flags = True
                flags = flags["or"]
            elif "and" in flags:
                flags = flags["and"]
            self.flags = flags
        else:
            self.flags = None
            self.or_flags = False
        if "values" in cfg:
            self.values = cfg["values"]
        else:
            self.values = None
        if not self.flags and not self.values:
            raise ConfigException(f"Value map rule in style {style_def.name} of layer {style_def.product.name} must have a non-empty 'flags' or a 'values' section.")
        if self.flags and self.values:
            raise ConfigException(f"Value map rule in style {style_def.name} of layer {style_def.product.name} has a both a 'flags' and a 'values' section - choose one.")

    def create_mask(self, data):
        if self.values:
            mask = None
            for v in self.values:
                vmask = data == v
                if mask is None:
                    mask = vmask
                else:
                    mask |= vmask
        elif self.or_flags:
            mask = None
            for f in self.flags.items():
                f = {f[0]: f[1]}
                if mask is None:
                    mask = make_mask(data, **f)
                else:
                    mask |= make_mask(data, **f)
        else:
            mask = make_mask(data, **self.flags)
        return mask

    @classmethod
    def value_map_from_config(cls, style, cfg):
        vmap = {}
        for band_name, rules in cfg.items():
            band_rules = [cls(style, band_name, rule) for rule in rules]
            vmap[band_name] = band_rules
        return vmap


class ColorMapStyleDef(StyleDefBase):
    auto_legend = True

    def __init__(self, product, style_cfg, stand_alone=False, user_defined=False):
        super().__init__(product, style_cfg, stand_alone=stand_alone, user_defined=user_defined)
        style_cfg = self._raw_cfg
        self.value_map = ValueMapRule.value_map_from_config(self, style_cfg["value_map"])
        for band in self.value_map.keys():
            self.raw_needed_bands.add(band)

    @staticmethod
    def reint(data):
        inted = data.astype("int")
        if hasattr(data, "attrs"):
            attrs = data.attrs
            inted.attrs = attrs
        return inted

    @staticmethod
    def create_colordata(data, rgb, alpha, mask):
        target = Dataset(coords=data.coords)
        colors = ["red", "green", "blue", "alpha"]
        for color in colors:
            val = alpha if color == "alpha" else getattr(rgb, color)
            c = numpy.full(data.shape, val)
            target[color] = DataArray(c, dims=data.dims, coords=data.coords)
        masked = target.where(mask).where(numpy.isfinite(data))  # remask
        return masked

    def transform_single_date_data(self, data):
        # pylint: disable=too-many-locals, too-many-branches
        # extent mask data per band to preseve nodata
        _LOG.debug("transform begin %s", datetime.now())
        # if extent_mask is not None:
        #    for band in data.data_vars:
        # try:
        #            data[band] = data[band].where(extent_mask, other=data[band].attrs['nodata'])
        #        except AttributeError:
        #            data[band] = data[band].where(extent_mask)

        imgdata = Dataset(coords=data.coords)
        for cfg_band, rules in self.value_map.items():
            # Run through each item
            band = self.product.band_idx.band(cfg_band)
            band_data = Dataset()
            bdata = data[band]
            if bdata.dtype.kind == 'f':
                # Convert back to int for bitmasking
                bdata = ColorMapStyleDef.reint(bdata)
            for rule in rules:
                mask = rule.create_mask(bdata)

                masked = ColorMapStyleDef.create_colordata(bdata, rule.rgb, rule.alpha, mask)
                band_data = masked if len(band_data.data_vars) == 0 else band_data.combine_first(masked)

            imgdata = band_data if len(imgdata.data_vars) == 0 else merge([imgdata, band_data])

        imgdata *= 255
        return imgdata.astype('uint8')

    def single_date_legend(self, bytesio):
        patches = []
        for band in self.value_map.keys():
            for rule in reversed(self.value_map[band]):
                # only include values that are not transparent (and that have a non-blank title or abstract)
                if rule.alpha > 0.001 and rule.label:
                    try:
                        patch = mpatches.Patch(color=rule.rgb.hex_l, label=rule.label)
                    # pylint: disable=broad-except
                    except Exception as e:
                        print("Error creating patch?", e)
                    patches.append(patch)
        cfg = self.legend_cfg
        plt.rcdefaults()
        if cfg.get("rcParams", None) is not None:
            plt.rcParams.update(cfg.get("rcParams"))
        figure = plt.figure(figsize=(cfg.get("width", 3),
                                     cfg.get("height", 1.25)))
        plt.axis('off')
        legend = plt.legend(handles=patches, loc='center', frameon=False)
        plt.savefig(bytesio, format='png')


StyleDefBase.register_subclass(ColorMapStyleDef, "value_map")



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
from xarray import Dataset

from datacube_ows.styles.base import StyleDefBase
from datacube_ows.styles.component import ComponentStyleDef
from datacube_ows.styles.ramp import ColorRampDef


class HybridStyleDef(ColorRampDef, ComponentStyleDef):
    auto_legend = False
    def __init__(self, product, style_cfg, defer_multi_date=False, stand_alone=False, user_defined=False):
        super(HybridStyleDef, self).__init__(product, style_cfg,
                                             defer_multi_date=defer_multi_date,
                                             stand_alone=stand_alone,
                                             user_defined=user_defined)
        style_cfg = self._raw_cfg
        self.component_ratio = style_cfg["component_ratio"]

    def transform_single_date_data(self, data):
        #pylint: disable=too-many-locals
        if self.index_function is not None:
            data['index_function'] = (data.dims, self.index_function(data).data)

        imgdata = Dataset(coords=data)

        d = data['index_function']
        for band, intensity in self.rgb_components.items():
            rampdata = self.color_ramp.get_value(d, band)
            component_band_data = None
            if band in self.rgb_components:
                for c_band, c_intensity in self.rgb_components[band].items():
                    if callable(c_intensity):
                        imgband_component_data = c_intensity(data[c_band], c_band, band)
                    else:
                        imgband_component_data = data[c_band] * c_intensity
                    if component_band_data is not None:
                        component_band_data += imgband_component_data
                    else:
                        component_band_data = imgband_component_data
                    if band != "alpha":
                        component_band_data = self.compress_band(band, component_band_data)
                img_band_data = (rampdata * 255.0 * (1.0 - self.component_ratio)
                                 + self.component_ratio * component_band_data)
            else:
                img_band_data = rampdata * 255.0
            imgdata[band] = (d.dims, img_band_data.astype("uint8").data)

        return imgdata


StyleDefBase.register_subclass(HybridStyleDef, "component_ratio", priority=True)



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
import lark


def formula_parser():
    return lark.Lark("""
                ?expr: num_expr | bool_expr

                ?bool_expr: or_clause | comparison_clause

                ?or_clause: or_clause "|" and_clause -> or_
                          | or_clause "^" and_clause -> xor
                          | and_clause
                ?and_clause: and_clause "&" term -> and_
                           | term
                ?term: "not" term -> not_
                     | "(" bool_expr ")"

                ?comparison_clause: eq | ne | le | ge | lt | gt

                eq: num_expr "==" num_expr
                ne: num_expr "!=" num_expr
                le: num_expr "<=" num_expr
                ge: num_expr ">=" num_expr
                lt: num_expr "<" num_expr
                gt: num_expr ">" num_expr

                ?num_expr: shift

                ?shift: shift "<<" sum -> lshift
                      | shift ">>" sum -> rshift
                      | sum

                ?sum: sum "+" product -> add
                    | sum "-" product -> sub
                    | product

                ?product: product "*" atom -> mul
                        | product "/" atom -> truediv
                        | product "//" atom -> floordiv
                        | product "%" atom -> mod
                        | atom

                ?atom: "-" subatom -> neg
                     | "+" subatom -> pos
                     | "~" subatom -> inv
                     | subatom "**" atom -> pow
                     | subatom

                ?subatom: NAME -> var_name
                        | FLOAT -> float_literal
                        | INT -> int_literal
                        | "(" num_expr ")"


                %import common.FLOAT
                %import common.INT
                %import common.WS_INLINE
                %import common.CNAME -> NAME

                %ignore WS_INLINE
                """, start='expr')



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
import lark

from datacube_ows.ogc_utils import ConfigException
from datacube_ows.styles.expr import formula_parser

identity = lambda ev, x: x

def empty_gen(ev, a):
    return set()

def union(ev, a, b):
    return a.union(b)

def not_supported(op_name):
    def impl(ev, a=None, b=None, c=None):
        raise ConfigException(f"{op_name} not supported")
    return impl


@lark.v_args(inline=True)
class ExpressionEvaluator(lark.Transformer):
    from operator import add, floordiv, mod, mul, neg, pos, pow, sub, truediv
    not_ = inv = or_ = and_ = xor = not_supported("Bitwise logical operators")
    eq = ne = le = ge = lt = gt = not_supported("Comparison operators")
    lshift = rshift = not_supported("Left and right-shift operators")

    float_literal = float
    int_literal = int

    def __init__(self, style, *args, **kwargs):
        self.ows_style = style
        super().__init__(*args, **kwargs)


@lark.v_args(inline=True)
class UserDefinedExpressionEvaluator(ExpressionEvaluator):
    pow = not_supported("Exponent operator")


@lark.v_args(inline=True)
class BandListEvaluator(ExpressionEvaluator):
    neg = pos = identity
    add = sub = mul = truediv = floordiv = mod = pow = union

    float_literal = empty_gen
    int_literal = empty_gen

    def var_name(self, key):
        return set([self.ows_style.local_band(key.value)])


class ExpressionException(ConfigException):
    pass

class Expression:
    def __init__(self, style, expr_str):
        self.style = style
        self.expr_str = expr_str
        parser = formula_parser()
        try:
            self.tree = parser.parse(self.expr_str)
            self.needed_bands = BandListEvaluator(self.style).transform(self.tree)
        except lark.LarkError as e:
            raise ExpressionException(f"Invalid expression: {e} {self.expr_str}")
        except KeyError as e:
            raise ExpressionException(f"Unrecognised band '{e}' in {expr_str}")
        if len(self.needed_bands) == 0:
            raise ExpressionException(f"Expression references no bands: {self.expr_str}")


    def __call__(self, data):
        if self.style.user_defined:
            evaluator = UserDefinedExpressionEvaluator
        else:
            evaluator = ExpressionEvaluator

        @lark.v_args(inline=True)
        class ExpressionDataEvaluator(evaluator):
            def var_name(self, key):
                return data[self.ows_style.local_band(key.value)]

        return ExpressionDataEvaluator(self.style).transform(self.tree)




# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
import numpy as np
from xarray import DataArray, Dataset

from datacube_ows.ogc_utils import ConfigException, FunctionWrapper
from datacube_ows.styles.base import StyleDefBase

# pylint: disable=abstract-method


class ComponentStyleDef(StyleDefBase):
    def __init__(self, product, style_cfg,
                 stand_alone=False, defer_multi_date=False, user_defined=False):
        super().__init__(product, style_cfg,
                         stand_alone=stand_alone, defer_multi_date=defer_multi_date, user_defined=user_defined)
        style_cfg = self._raw_cfg
        self.raw_rgb_components = {}
        for imgband in ["red", "green", "blue", "alpha"]:
            components = style_cfg["components"].get(imgband)
            if components is None:
                if imgband == "alpha":
                    continue
                else:
                    raise ConfigException(f"No components defined for {imgband} band in style {self.name}, layer {product.name}")
            if callable(components) or "function" in components:
                self.raw_rgb_components[imgband] = FunctionWrapper(self.product, components,
                                                                   stand_alone=self.stand_alone)
                if not self.stand_alone:
                    if "additional_bands" not in style_cfg:
                        raise ConfigException(f"Style with a function component must declare additional_bands.")
                    for b in style_cfg.get("additional_bands", set()):
                        self.raw_needed_bands.add(b)
            else:
                self.raw_rgb_components[imgband] = components
                for k in components.keys():
                    if k != "scale_range":
                        self.raw_needed_bands.add(k)
        self.declare_unready("rgb_components")

        self.scale_factor = style_cfg.get("scale_factor")
        if "scale_range" in style_cfg:
            self.scale_min, self.scale_max = style_cfg["scale_range"]
        elif self.scale_factor:
            self.scale_min = 0.0
            self.scale_max = 255.0 * self.scale_factor
        else:
            self.scale_min = None
            self.scale_max = None

        self.component_scale_ranges = {}
        for cn, cd in style_cfg["components"].items():
            if not callable(cd) and "scale_range" in cd:
                self.component_scale_ranges[cn] = {
                    "min": cd["scale_range"][0],
                    "max": cd["scale_range"][1],
                }
            else:
                self.component_scale_ranges[cn] = {
                    "min": self.scale_min,
                    "max": self.scale_max,
                }

    # pylint: disable=attribute-defined-outside-init
    def make_ready(self, dc, *args, **kwargs):
        self.rgb_components = {}
        for band, component in self.raw_rgb_components.items():
            if not component or callable(component):
                self.rgb_components[band] = component
            else:
                self.rgb_components[band] = self.dealias_components(component)
        super().make_ready(dc, *args, **kwargs)

    def dealias_components(self, comp_in):
        if self.stand_alone:
            return comp_in
        elif comp_in is None:
            return None
        else:
            return {self.product.band_idx.band(self.local_band(band_alias)): value for band_alias, value in comp_in.items() if band_alias not in ['scale_range']}

    def compress_band(self, component_name, imgband_data):
        sc_min = self.component_scale_ranges[component_name]["min"]
        sc_max = self.component_scale_ranges[component_name]["max"]
        clipped = imgband_data.clip(sc_min, sc_max)
        normalized = (clipped - sc_min) / (sc_max - sc_min)
        return normalized * 255


    def transform_single_date_data(self, data):
        imgdata = {}
        for imgband, components in self.rgb_components.items():
            if callable(components):
                imgband_data = components(data)
                dims = imgband_data.dims
                imgband_data = imgband_data.astype('uint8')
                imgdata[imgband] = imgband_data
            else:
                imgband_data = None
                for band, intensity in components.items():
                    if callable(intensity):
                        imgband_component = intensity(data[band], band, imgband)
                    elif band == "scale_range":
                        continue
                    else:
                        imgband_component = data[band] * intensity

                    if imgband_data is not None:
                        imgband_data += imgband_component
                    else:
                        imgband_data = imgband_component
                if imgband_data is None:
                    imgband_data = np.zeros(list(data.dims.values()), 'uint8')
                    imgband_data = DataArray(imgband_data, data.coords, data.dims.keys())
                if imgband != "alpha":
                    imgband_data = self.compress_band(imgband, imgband_data)
                imgdata[imgband] = imgband_data.astype("uint8")

        image_dataset = Dataset(imgdata)
        return image_dataset


StyleDefBase.register_subclass(ComponentStyleDef, "components")



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import absolute_import, division, print_function

import traceback as tb

from flask import render_template
from ows.common.types import OWSException, Version
from ows.common.v20.encoders import xml_encode_exception_report

from datacube_ows.ogc_utils import resp_headers


class OGCException(Exception):
    INVALID_FORMAT = "InvalidFormat"
    CURRENT_UPDATE_SEQUENCE = "CurrentUpdateSequence"
    INVALID_UPDATE_SEQUENCE = "InvalidUpdateSequence"

    version = None
    schema_url = None

    # pylint: disable=super-init-not-called
    def __init__(self, msg, code=None, locator=None, http_response=400, valid_keys=None):
        self.http_response = http_response
        self.errors = []
        self.add_error(msg, code, locator, valid_keys)

    def add_error(self, msg, code=None, locator=None, valid_keys=None):
        self.errors.append({
            "msg": msg,
            "code": code,
            "locator": locator,
            "valid_keys": valid_keys
        })

    # pylint: disable=dangerous-default-value
    def exception_response(self, traceback=[]):
        return (render_template("ogc_error.xml",
                                exception=self,
                                traceback=traceback,
                                version=self.version,
                                schema_url=self.schema_url),
                self.http_response,
                resp_headers({"Content-Type": "application/xml"})
               )


class WMSException(OGCException):
    INVALID_CRS = "InvalidCRS"
    LAYER_NOT_DEFINED = "LayerNotDefined"
    STYLE_NOT_DEFINED = "StyleNotDefined"
    LAYER_NOT_QUERYABLE = "LayerNotQueryable"
    INVALID_POINT = "InvalidPoint"
    MISSING_DIMENSION_VALUE = "MissingDimensionValue"
    INVALID_DIMENSION_VALUE = "InvalidDimensionValue"
    OPERATION_NOT_SUPPORTED = "OperationNotSupported"

    version = "1.3.0"
    schema_url = "http://schemas.opengis.net/wms/1.3.0/exceptions_1_3_0.xsd"


class WMTSException(WMSException):
    version = "1.0.0"
    schema_url = "http://schemas.opengis.net/ows/1.1.0/owsExceptionReport.xsd"


class WCS1Exception(OGCException):
    COVERAGE_NOT_DEFINED = "CoverageNotDefined"
    MISSING_PARAMETER_VALUE = "MissingParameterValue"
    INVALID_PARAMETER_VALUE = "InvalidParameterValue"

    version = "1.2.0"
    schema_url = "http://schemas.opengis.net/wcs/1.0.0/OGC-exception.xsd"


class WCS2Exception(OGCException):
    NO_SUCH_COVERAGE = "NoSuchCoverage"
    INVALID_AXIS_LABEL = "InvalidAxisLabel"
    INVALID_SUBSETTING = "InvalidSubsetting"
    EMPTY_COVERAGE_ID_LIST = "emptyCoverageIdList"
    NO_SUCH_FIELD = "NoSuchField"
    ILLEGAL_FIELD_SEQUENCE = "IllegalFieldSequence"
    INVALID_COVERAGE_TYPE = "InvalidCoverageType"
    INVALID_SCALE_FACTOR = "InvalidScaleFactor"
    INVALID_EXTENT = "InvalidExtent"
    SCALE_AXIS_UNDEFINED = "ScaleAxisUndefined"
    NOT_A_CRS = "NotACrs"
    CRS_MISMATCH = "CrsMismatch"
    SUBSETTING_CRS_NOT_SUPPORTED = "SubsettingCrsNotSupported"
    OUTPUT_CRS_NOT_SUPPORTED = "OutputCrsNotSupported"
    INTERPOLATION_METHOD_NOT_SUPPORTED = "InterpolationMethodNotSupported"
    NO_SUCH_AXIS = "NoSuchAxis"
    MISSING_PARAMETER_VALUE = "MissingParameterValue"
    INVALID_PARAMETER_VALUE = "InvalidParameterValue"

    version = "2.0.1"
    schema_url = "http://schemas.opengis.net/wcs/2.0/wcsAll.xsd"

    # pylint: disable=dangerous-default-value
    def exception_response(self, traceback=[]):
        exceptions = [
            OWSException(
                code=error['code'],
                locator=error['locator'],
                text=[error['msg']] + tb.format_list(traceback)
            )
            for error in self.errors
        ]

        if traceback:
            exceptions[0].traceback = tb.format_list(traceback)

        result = xml_encode_exception_report(exceptions, Version.from_str(self.version))
        return (
            result.value,
            self.http_response,
            resp_headers({"Content-Type": "application/xml"})
        )



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
from contextlib import contextmanager
from threading import Lock
from typing import Generator, MutableMapping, Optional

from datacube import Datacube

_LOG: logging.Logger = logging.getLogger(__name__)

# CubePool class
class CubePool:
    """
    A Cube pool is a thread-safe resource pool for managing Datacube objects (which map to database connections).
    """
    # _instances, global mapping of CubePools by app name
    _instances: MutableMapping[str, "CubePool"] = {}

    _cubes_: bool = False
    _cubes_lock_: bool = False

    def __new__(cls, app: str) -> "CubePool":
        """
        Construction of CubePools is managed. Constructing a cubepool for an app string that already has a cubepool
        constructed, returns the existing cubepool, not a new one.
        """
        if app not in cls._instances:
            cls._instances[app] = super(CubePool, cls).__new__(cls)
        return cls._instances[app]

    def __init__(self, app: str) -> None:
        """
        Obtain the cube pool for the nominated app string, or create one if one does not exist yet.

        :param app: The app string used to construct any Datacube objects created by the pool.
        """
        self.app: str = app
        if not self._cubes_:
            self._cubes: MutableMapping[Datacube, bool] = {}
            self._cubes_ = True
        if not self._cubes_lock_:
            self._cubes_lock: Lock = Lock()
            self._cubes_lock_ = True

    def get_cube(self) -> Optional[Datacube]:
        """
        Return a Datacube object.  Either generating a new Datacube, or recycling an unassigned one already in the pool.

        :return:  a Datacube object (or None on error).
        """
        self._cubes_lock.acquire()
        for c, assigned in self._cubes.items():
            if not assigned:
                self._cubes[c] = True
                self._cubes_lock.release()
                return c
        try:
            c = self._new_cube()
            self._cubes[c] = True
        # pylint: disable=broad-except
        except Exception as e:
            _LOG.error("ODC initialisation failed: %s", str(e))
            c = None
        finally:
            self._cubes_lock.release()
        return c

    def release_cube(self, c: Datacube) -> None:
        """
        Return a datacube to the pool for reassignment.

        :param c: A Datacube object originally created by this CubePool and that is no longer required.
        """
        if c not in self._cubes:
            raise Exception("Cannot release non-pool datacube.")
        self._cubes_lock.acquire()
        self._cubes[c] = False
        self._cubes_lock.release()

    def _new_cube(self) -> Datacube:
        return Datacube(app=self.app)

    def __len__(self) -> int:
        return len(self._cubes)


# Lowlevel CubePool API
def get_cube(app: str = "ows") -> Optional[Datacube]:
    """
    Obtain a Datacube object from the appropriate pool

    :param app: The app pool to use - defaults to "ows".
    :return: a Datacube object (or None) in case of database error.
    """
    pool = CubePool(app=app)
    return pool.get_cube()


def release_cube(c: Datacube, app: str = "ows") -> None:
    """
    Return a Datacube to the pool for reuse.

    :param c: a Datacube object, allocated by a previous call to get_cube(app).
    :param app: the name of the pool to return the cube from - defaults to "ows"
    """
    pool = CubePool(app=app)
    return pool.release_cube(c)


def pool_size(app: str = "ows") -> int:
    """
    Return the total number of cubes (available and allocated) in a particular pool.

    :param app: The name of the pool to measure - defaults to "ows"
    :return: the total number of cubes in the pool.
    """
    pool = CubePool(app=app)
    return len(pool)


# High Level Cube Pool API
@contextmanager
def cube(app: str = "ows") -> Generator[Optional["datacube.api.core.Datacube"], None, None]:
    """
    Context manager for using a Datacube object from a pool.

    E.g.

    with cube() as dc:
         # Do stuff that needs a datacube object here.
         data = dc.load(.....)


    :param app: The pool to obtain the app from - defaults to "ows".
    :return: A Datacube context manager.
    """
    dc = get_cube(app)
    try:
        yield dc
    finally:
        if dc:
            release_cube(dc, app)



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import absolute_import, division, print_function

import collections
import logging

import datacube
from datacube.utils import geometry
from dateutil.parser import parse
from ows.wcs.v20 import ScaleAxis, ScaleExtent, ScaleSize, Slice, Trim
from rasterio import MemoryFile

from datacube_ows.cube_pool import cube
from datacube_ows.data import DataStacker
from datacube_ows.mv_index import MVSelectOpts
from datacube_ows.ogc_exceptions import WCS2Exception
from datacube_ows.ows_configuration import get_config
from datacube_ows.resource_limits import ResourceLimited
from datacube_ows.wcs_scaler import WCSScaler, WCSScalerUnknownDimension

_LOG = logging.getLogger(__name__)


def uniform_crs(cfg, crs):
    " Helper function to transform a URL style EPSG definition to an 'EPSG:nnn' one "
    if crs.startswith('http://www.opengis.net/def/crs/EPSG/'):
        code = crs.rpartition('/')[-1]
        crs = 'EPSG:%s' % code
    elif crs.startswith('urn:ogc:def:crs:EPSG:'):
        code = crs.rpartition(':')[-1]
        crs = 'EPSG:%s' % code
    elif crs.startswith('EPSG'):
        pass
    elif crs in cfg.published_CRSs:
        pass
    else:
        raise WCS2Exception("Not a CRS: %s" % crs,
                            WCS2Exception.NOT_A_CRS,
                            locator=crs,
                            valid_keys=list(cfg.published_CRSs))
    return crs


def get_coverage_data(request):
    # pylint: disable=too-many-locals, protected-access

    cfg = get_config()

    layer_name = request.coverage_id
    layer = cfg.product_index.get(layer_name)
    if not layer or not layer.wcs:
        raise WCS2Exception("Invalid coverage: %s" % layer_name,
                            WCS2Exception.NO_SUCH_COVERAGE,
                            locator="COVERAGE parameter",
                            valid_keys=list(cfg.product_index))

    with cube() as dc:
        if not dc:
            raise WCS2Exception("Database connectivity failure")
        #
        # CRS handling
        #

        native_crs = layer.native_CRS
        subsetting_crs = uniform_crs(cfg, request.subsetting_crs or native_crs)
        output_crs = uniform_crs(cfg, request.output_crs or subsetting_crs)

        if subsetting_crs not in cfg.published_CRSs:
            raise WCS2Exception("Invalid subsettingCrs: %s" % subsetting_crs,
                                WCS2Exception.SUBSETTING_CRS_NOT_SUPPORTED,
                                locator=subsetting_crs,
                                valid_keys=list(cfg.published_CRSs))

        output_crs = uniform_crs(cfg, request.output_crs or subsetting_crs or native_crs)

        if output_crs not in cfg.published_CRSs:
            raise WCS2Exception("Invalid outputCrs: %s" % output_crs,
                                WCS2Exception.OUTPUT_CRS_NOT_SUPPORTED,
                                locator=output_crs,
                                valid_keys=list(cfg.published_CRSs))

        #
        # Subsetting/Scaling
        #

        scaler = WCSScaler(layer, subsetting_crs)
        times = layer.ranges["times"]

        subsets = request.subsets

        if len(subsets) != len(set(subset.dimension.lower() for subset in subsets)):
            dimensions = [subset.dimension.lower() for subset in subsets]
            duplicate_dimensions = [
                item
                for item, count in collections.Counter(dimensions).items()
                if count > 1
            ]

            raise WCS2Exception("Duplicate dimension%s: %s" % (
                                    's' if len(duplicate_dimensions) > 1 else '',
                                    ', '.join(duplicate_dimensions)
                                ),
                                WCS2Exception.INVALID_SUBSETTING,
                                locator=','.join(duplicate_dimensions)
                                )

        for subset in subsets:
            dimension = subset.dimension.lower()
            if dimension == 'time':
                if isinstance(subset, Trim):
                    low = parse(subset.low).date() if subset.low is not None else None
                    high = parse(subset.high).date() if subset.high is not None else None
                    if low is not None:
                        times = [
                            time for time in times
                            if time >= low
                        ]
                    if high is not None:
                        times = [
                            time for time in times
                            if time <= high
                        ]
                elif isinstance(subset, Slice):
                    point = parse(subset.point).date()
                    times = [point]

            else:
                try:
                    if isinstance(subset, Trim):
                        scaler.trim(dimension, subset.low, subset.high)
                    elif isinstance(subset, Slice):
                        scaler.slice(dimension, subset.point)
                except WCSScalerUnknownDimension:
                    raise WCS2Exception('Invalid subsetting axis %s' % subset.dimension,
                                    WCS2Exception.INVALID_AXIS_LABEL,
                                    locator=subset.dimension)

        #
        # Transform spatial extent to native CRS.
        #
        scaler.to_crs(output_crs)

        #
        # Scaling
        #

        scales = request.scales
        if len(scales) != len(set(subset.axis.lower() for subset in scales)):
            axes = [subset.axis.lower() for subset in scales]
            duplicate_axes = [
                item
                for item, count in collections.Counter(axes).items()
                if count > 1
            ]
            raise WCS2Exception('Duplicate scales for ax%ss: %s' % (
                                    'i' if len(duplicate_axes) == 1 else 'e',
                                    ', '.join(duplicate_axes)
                                ),
                                WCS2Exception.INVALID_SCALE_FACTOR,
                                locator=','.join(duplicate_axes)
                                )

        for scale in scales:
            axis = scale.axis.lower()

            if axis in ('time', 'k'):
                raise WCS2Exception('Cannot scale axis %s' % scale.axis,
                                    WCS2Exception.INVALID_SCALE_FACTOR,
                                    locator=scale.axis
                                    )
            else:
                if isinstance(scale, ScaleAxis):
                    scaler.scale_axis(axis, scale.factor)
                elif isinstance(scale, ScaleSize):
                    scaler.scale_size(axis, scale.size)
                elif isinstance(scale, ScaleExtent):
                    scaler.scale_extent(axis, scale.low, scale.high)

        #
        # Rangesubset
        #

        band_labels = layer.band_idx.band_labels()
        if request.range_subset:
            bands = []
            for range_subset in request.range_subset:
                if isinstance(range_subset, str):
                    if range_subset not in band_labels:
                        raise WCS2Exception('No such field %s' % range_subset,
                                    WCS2Exception.NO_SUCH_FIELD,
                                    locator=range_subset,
                                    valid_keys=band_labels
                                    )
                    bands.append(range_subset)
                else:
                    if range_subset.start not in band_labels:
                        raise WCS2Exception('No such field %s' % range_subset.start,
                                            WCS2Exception.ILLEGAL_FIELD_SEQUENCE,
                                            locator=range_subset.start,
                                            valid_keys = band_labels)
                    if range_subset.end not in band_labels:
                        raise WCS2Exception('No such field %s' % range_subset.end,
                                            WCS2Exception.ILLEGAL_FIELD_SEQUENCE,
                                            locator=range_subset.end,
                                            valid_keys = band_labels)

                    start = band_labels.index(range_subset.start)
                    end = band_labels.index(range_subset.end)
                    bands.extend(band_labels[start:(end + 1) if end > start else (end - 1)])
        else:
            bands = layer.wcs_default_bands  # TODO: standard says differently

        #
        # Format handling
        #

        if not request.format:
            fmt = cfg.wcs_formats_by_name[layer.native_format]
        else:
            try:
                fmt = cfg.wcs_formats_by_mime[request.format]
            except KeyError:
                raise WCS2Exception("Unsupported format: %s" % request.format,
                                    WCS2Exception.INVALID_PARAMETER_VALUE,
                                    locator="FORMAT",
                                    valid_keys=list(cfg.wcs_formats_by_mime))

        if len(times) > 1 and not fmt.multi_time:
            raise WCS2Exception(
                "Format does not support multi-time datasets - "
                "either constrain the time dimension or choose a different format",
                WCS2Exception.INVALID_SUBSETTING,
                locator="FORMAT or SUBSET"
                                )
        affine = scaler.affine()
        geobox = geometry.GeoBox(scaler.size.x, scaler.size.y,
                                 affine, cfg.crs(output_crs))

        stacker = DataStacker(layer,
                              geobox,
                              times,
                              bands=bands)
        n_datasets = stacker.datasets(dc.index, mode=MVSelectOpts.COUNT)

        if n_datasets == 0:
            raise WCS2Exception("The requested spatio-temporal subsets return no data.",
                            WCS2Exception.INVALID_SUBSETTING,
                            http_response=404)

        try:
            layer.resource_limits.check_wcs(n_datasets)
        except ResourceLimited as e:
            raise WCS2Exception(
                f"This request processes too much data to be served in a reasonable amount of time. ({e}) "
                + "Please reduce the bounds of your request and try again.")

        datasets = stacker.datasets(dc.index)
        if fmt.multi_time and len(times) > 1:
            # Group by solar day
            group_by = datacube.api.query.query_group_by(time=times, group_by='solar_day')
            datasets = dc.group_datasets(datasets, group_by)

        output = stacker.data(datasets, skip_corrections=True)

        # Clean extent flag band from output
        raw_bands = [layer.band_idx.locale_band(b) for b in bands]
        for k, v in output.data_vars.items():
            if k not in raw_bands:
                output = output.drop_vars([k])

    #
    # TODO: configurable
    #
    if fmt.mime == 'image/geotiff':
        output = fmt.renderer(request.version)(request, output, output_crs,
                              layer, scaler.size.x, scaler.size.y, affine)

    else:
        output = fmt.renderer(request.version)(request, output, output_crs)

    headers = {
        "Content-Type": fmt.mime,
        'content-disposition': f'attachment; filename={request.coverage_id}.{fmt.extension}',
    }
    headers.update(layer.resource_limits.wcs_cache_rules.cache_headers(n_datasets))
    return output, headers


def get_tiff(request, data, crs, product, width, height, affine):
    """Uses rasterio MemoryFiles in order to return a streamable GeoTiff response"""
    # Copied from CEOS.  Does not seem to support multi-time dimension data - is this even possible in GeoTiff?
    supported_dtype_map = {
        'uint8': 1,
        'uint16': 2,
        'int16': 3,
        'uint32': 4,
        'int32': 5,
        'float32': 6,
        'float64': 7,
        'complex': 9,
        'complex64': 10,
        'complex128': 11,
    }

    dtype_list = [data[array].dtype for array in data.data_vars]
    dtype = str(max(dtype_list, key=lambda d: supported_dtype_map[str(d)]))

    # TODO: convert other parameters as-well
    gtiff = request.geotiff_encoding_parameters
    cfg = get_config()

    data = data.squeeze(dim="time", drop=True)
    data = data.astype(dtype)
    nodata = 0
    for band in data.data_vars:
        nodata = product.band_idx.nodata_val(band)
    with MemoryFile() as memfile:
        # pylint: disable=protected-access, bad-continuation

        kwargs = {}
        if gtiff.tile_width is not None:
            kwargs['blockxsize'] = gtiff.tile_width
        if gtiff.tile_height is not None:
            kwargs['blockysize'] = gtiff.tile_height

        if gtiff.predictor:
            predictor = gtiff.predictor.lower()
            if predictor == 'horizontal':
                kwargs['predictor'] = 2
            elif predictor == 'floatingpoint':
                kwargs['predictor'] = 3

        with memfile.open(
            driver="GTiff",
            width=width,
            height=height,
            count=len(data.data_vars),
            transform=affine,
            crs=crs,
            nodata=nodata,
            tiled=gtiff.tiling if gtiff.tiling is not None else True,
            compress=gtiff.compression.lower() if gtiff.compression else "lzw",
            predictor=2,
            interleave=gtiff.interleave or "band",
            dtype=dtype, **kwargs) as dst:
            for idx, band in enumerate(data.data_vars, start=1):
                dst.write(data[band].values, idx)
                dst.set_band_description(idx, product.band_idx.band_label(band))
                if cfg.wcs_tiff_statistics:
                    dst.update_tags(idx, STATISTICS_MINIMUM=data[band].values.min())
                    dst.update_tags(idx, STATISTICS_MAXIMUM=data[band].values.max())
                    dst.update_tags(idx, STATISTICS_MEAN=data[band].values.mean())
                    dst.update_tags(idx, STATISTICS_STDDEV=data[band].values.std())
        return memfile.read()


def get_netcdf(request, data, crs):
    # Cleanup dataset attributes for NetCDF export
    data.attrs["crs"] = crs # geometry.CRS(response_crs)
    for v in data.data_vars.values():
        v.attrs["crs"] = crs
        if "spectral_definition" in v.attrs:
            del v.attrs["spectral_definition"]
        if "flags_definition" in v.attrs:
            del v.attrs["flags_definition"]
    if "time" in data and "units" in data["time"].attrs:
        del data["time"].attrs["units"]

    # And export to NetCDF
    return data.to_netcdf()
